
"""
Argument Parsing:
* Script level.
    * This is system dependent and is hard in bash.
    * It remains a necessity because at system level one can dictate the enviroment, the interpretor, etc.
* Python level:
    * system agnostic.
    * Benign syntax, but predetermines to a great extent what is being executed (which python, enviroment).
* python library `fire`:
    * this is good for passing arguments to specfic python functions from commandline without writing specific argparsing for those functions.

The best approach is to use python and fire to pass process args of script and:
   * return a string to the script file to execute it.
   * or, execute it via terminal from within python.

Choices made by default:
* ipython over python
* interactive is the default
* importing the file to be run (as opposed to running it as main) is the default. The advantage of running it as an imported module is having reference to the file from which classes came. This is vital for pickling.

"""

# import crocodile.toolbox as tb
import argparse
import subprocess
import platform
from pathlib import Path


def build_parser():
    parser = argparse.ArgumentParser(description="Generic Parser to launch crocodile shell.")

    # POSITIONAL ARGUMENT (UNNAMED)
    # if dest is not specified, then, it has same path as keyword, e.g. "--dest"
    # parser.add_argument("--file", "-f", dest="file", help="Python file path.", default="")

    # A FLAG:
    parser.add_argument("--module", '-m', help="Flag tells to run the file as main.", action="store_true", default=False)  # default is running as main, unless indicated by --module flag.
    parser.add_argument("--newWindow", "-w", help="Flag for running in new window.", action="store_true", default=False)
    parser.add_argument("--nonInteratctive", "-N", help="Specify a non-interactive session.", action="store_true", default=False)
    parser.add_argument("--python", "-p", help="Use python over IPython.", action="store_true", default=False)

    # OPTIONAL KEYWORD
    parser.add_argument("--kill", "-k", dest="kill", help="Python file path.", default="")
    parser.add_argument("--file", "-f", dest="file", help="Python file path.", default="")
    parser.add_argument("--func", "-F", dest="func", help=f"function to be run after import", default="")
    parser.add_argument("--cmd", "-c", dest="cmd", help="Python command.", default="")
    parser.add_argument("--read", "-r", dest="read", help="Read file", default="")
    parser.add_argument("--terminal", "-t", dest="terminal",  help=f"Flag to specify which terminal to be used. Default console host.", default="")  # can choose `wt`
    parser.add_argument("--shell", "-S", dest="shell", help=f"Flag to specify which terminal to be used. Default CMD.", default="")

    args = parser.parse_args()
    # print(f"Crocodile.run: args of the firing command = {args.__dict__}")

    # ==================================================================================
    # flags processing
    interactivity = '' if args.nonInteratctive else '-i'
    interpreter = 'python' if args.python else 'ipython'

    if args.file != "":
        file = Path(args.file).expanduser().absolute()
        if not args.module: res = f"{interpreter} {interactivity} {file}"
        else:  # run as a module (i.e. import it)
            script = fr"""
from {file} import *
args.cmd"""
            if args.func != "": script += f"tb.E.capture_locals({args.func}, globals())"
            res = f"{interpreter} {interactivity} {script}"
    elif args.cmd != "":
        # res = f""" python -c "from crocodile.toolbox import *; import crocodile.environment as env; {args.cmd}" """
        import textwrap
        code = f"from crocodile.toolbox import *\n{textwrap.dedent(args.cmd)}"
        # print(code)
        exec(code)
        return None
    elif args.read != "":
        code_text = f"""
# >>>>>>> Reading File <<<<<<<<<
p = P(r\'{str(args.read).lstrip()}\').absolute()
try:
    dat = p.readit()
except Exception as e:
    print(e)
"""

        # next, write code_text to file at home directory /tmp_results/shells/python_readfile_script.py using open:
        code_file = Path.home().joinpath("tmp_results/shells/python_readfile_script.py")
        code_file.write_text(code_text)
        res = f"""ipython --no-banner -i -m crocodile.croshell -- --file "{code_file}" """
    elif args.kill != "":
        res = __import__("crocodile.toolbox").toolbox.L(__import__("psutil").process_iter()).filter(lambda x: args.kill in x.name())
        # res.print()
        for item in res[::-1]: print(f"killing {item.name()}"); item.kill()
        return None
    else:
        res = f"{interpreter} {interactivity} --no-banner -m crocodile.croshell"  # --term-title croshell
        # Clear-Host;
        # # --autocall 1 in order to enable shell-like behaviour: e.g.: P x is interpreted as P(x)

    # print(res)
    if platform.system() == "Windows": return subprocess.run([f"powershell", "-Command", res], shell=True, capture_output=False, text=True)
    else: return subprocess.run([res], shell=True, capture_output=False, text=True)


if __name__ == "__main__":
    build_parser()




import time
import crocodile.toolbox as tb
from rich.progress import track


def expensive_function() -> tb.P:
    execution_time_in_seconds = 60 * 1.2
    print(f"Hello, I am an expensive function, and I just started running ...\n It will take me {execution_time_in_seconds} seconds to finish")
    a = 1 + 1

    steps = 100
    for _ in track(range(steps), description="Progress bar ..."):
        time.sleep(execution_time_in_seconds/steps)  # Simulate work being done

    print("I'm done, I crunched a lot of numbers. Next I'll save my results to a file and passing its directory back to the main process on the machine running me.")
    path = tb.P.tmpdir().joinpath("result.Struct.pkl")
    tb.S(a=a).save(path=path)
    return path.parent  # extra job details files will be added to this res_folder


# parallelizeBegins


def parallelize(idx_start: int, idx_end: int, idx_max: int, num_instances: int) -> tb.P:

    print(f"This script will execute ({(idx_end - idx_start) / idx_max * 100:.2f}%) of the work on this machine.")

    print(f"Splitting the work ({idx_start=}, {idx_end=}) equally among {num_instances} instances via `parallelize` of cluster.trial_file ...")

    kwargs_split = tb.L(range(idx_start, idx_end, 1)).split(to=num_instances).apply(lambda sub_list: dict(idx_start=sub_list[0], idx_end=sub_list[-1], idx_max=idx_max))
    for idx, x in enumerate(kwargs_split):
        tb.S(x).print(as_config=True, title=f"Instance {idx}")
    print("\n" * 2)

    save_dir_suffix = f"machine_{idx_start}_{idx_end}"

    res = kwargs_split.apply(lambda kwargs: expensive_function_single_thread(**kwargs, save_dir_suffix=save_dir_suffix), jobs=num_instances)
    return tb.P(res[0]).parent


# parallelizeEnds


# todo: consider making it a method of a class and user will subclass it and is forced to follow its interface


def expensive_function_single_thread(idx_start: int, idx_end: int, idx_max: int, save_dir_suffix: tb.P or str) -> tb.P:
    print(f"Hello, I am one thread of an expensive function, and I just started running ...")
    execution_time_in_seconds = 60 * 1
    steps = 100
    for _ in track(range(steps), description="Progress bar ..."):
        time.sleep(execution_time_in_seconds/steps)  # Simulate work being done
    print("I'm done, I crunched numbers from {} to {}.".format(idx_start, idx_end))
    _ = idx_max
    save_dir = tb.P.tmp().joinpath(f"tmp_dirs/expensive_function_single_thread").joinpath(save_dir_suffix, f"thread_{idx_start}_{idx_end}").create()
    tb.S(a=1).save(path=save_dir.joinpath(f"trial_func_result.Struct.pkl"))
    return save_dir


if __name__ == '__main__':
    pass




"""
This module uses the generic google-api-python-client, which is very low level.

read: https://developers.google.com/drive/api/guides/search-files
file specs: https://developers.google.com/drive/api/v3/reference/files
# https://github.com/eduardogr/google-drive-python
# google: python gdrive api expiry date

# code from https://developers.google.com/drive/api/quickstart/python
terminology: https://developers.google.com/workspace/guides/auth-overview

Steps:
* Create a project on console.developers.google.com which redirects to https://console.cloud.google.com/projectselector2/apis/dashboard
* Enable the Drive API
* Create credentials for a web server to access application data
"""

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google.oauth2 import service_account
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
# from googleapiclient.errors import HttpError

from googleapiclient.http import MediaIoBaseDownload, MediaFileUpload

from crocodile.comms.helper_funcs import process_sent_file, process_retrieved_file
from enum import Enum
import crocodile.toolbox as tb
import io
import pandas as pd


tb.Display.set_pandas_display()


class Scopes(Enum):  # TODO: If modifying these scopes, delete the file token.json.
    admin =      tb.P('https://www.googleapis.com/auth/drive')  # See, edit, create, and delete all of your Google Drive files
    settings =   tb.P('https://www.googleapis.com/auth/drive.appdata')  # View and manage its own configuration data in your Google Drive
    manage =     tb.P('https://www.googleapis.com/auth/drive.file')  # View and manage Google Drive files and folders that you have opened or created with this app
    managemeta = tb.P('https://www.googleapis.com/auth/drive.metadata')  # View and manage metadata of files in your Google # Drive
    readmeta =   tb.P('https://www.googleapis.com/auth/drive.metadata.readonly')  # View metadata for files in your Google Drive
    photo =      tb.P('https://www.googleapis.com/auth/drive.photos.readonly')  # View the photos, videos and albums in your # Google Photos
    readonly =   tb.P('https://www.googleapis.com/auth/drive.readonly')  # See and download all your Google Drive files
    app =        tb.P('https://www.googleapis.com/auth/drive.scripts')  # Modify your Google Apps Script scripts' behavior


class GDriveAPI:
    def __init__(self, account=None, project=None, scopes=None):
        scopes = [Scopes.admin] if scopes is None else scopes
        self.SCOPES = [scope.value.as_url_str() for scope in scopes]
        self.creds = self.get_cred(account, project)
        self.service = build('drive', 'v3', credentials=self.creds)
        self.drive_url = tb.P("https://drive.google.com/drive/u/1")

    def get_cred(self, account, project):
        config = tb.P.home().joinpath("dotfiles/google/drive/config.toml").readit()
        if account is None: account = list(config.keys())[0]; print(f"GDRIVE: using default account `{account}`")
        if project is None: project = list(config[account].keys())[0]; print(f"GDRIVE: using default project `{project}`")
        api_key = config[account][project]['api_key']
        client_id_file = tb.P.home().joinpath(f"dotfiles/google/drive/{config[account][project]['auth_client_config']}")
        creds = None  # The file token.json stores the user's access and refresh tokens, and is created automatically when the authorization flow completes for the first time.

        if client_id_file.exists():
            # client_id_info = client_id_file.readit()
            # if "refresh_token" in client_id_info:  # refersh token is not available before at least the first use.
            creds = Credentials.from_authorized_user_file(filename=client_id_file, scopes=self.SCOPES)

        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:  # If there are no (valid) credentials available, let the user log in.
            if creds and creds.expired and creds.refresh_token:
                try: creds.refresh(Request())
                except Exception: creds = self.authorize_screen()
            else: creds = self.authorize_screen()
            client_id_file.write_text(creds.to_json())
        return creds

    def authorize_screen(self):
        auth_secret = tb.P.home().joinpath(f"dotfiles/google/drive/Oauth_client_secret.json")
        flow = InstalledAppFlow.from_client_secrets_file(client_secrets_file=auth_secret, scopes=self.SCOPES)
        creds = flow.run_local_server(port=0)
        return creds

    def search_all(self, q='', size=1000): return pd.DataFrame(self.service.files().list(q=q, pageSize=size, fields="nextPageToken, files(id, name)", spaces="drive").execute().get('files', []), columns=['id', 'name']).set_index("name")
    def search_folders(self): return pd.DataFrame(self.service.files().list(q="mimeType = 'application/vnd.google-apps.folder'").execute()['files'])
    def get_fields_from_id(self, fid, fields="*") -> dict: return self.service.files().get(fileId=fid, fields=fields).execute()  # e.g: fields="id, name, mimeType, size, modifiedTime, createdTime, trashed, parents"
    def get_path_from_id(self, fid) -> str:
        tmp = self.get_fields_from_id(fid, 'name, parents')
        name = tmp['name']
        parent_id = tmp['parents'][0] if 'parents' in tmp else None
        if parent_id is None: return name  # name will be 'My Drive'
        else: return self.get_path_from_id(parent_id) + "/" + name

    def get_id_from_path(self, path) -> str:
        parent_id = 'root'
        for item in tb.P(path).items:
            tmp = self.service.files().list(q=f"(name={repr(item)}) and ({repr(parent_id)} in parents)", fields="nextPageToken, files(kind, id, name, parents)").execute()['files']
            assert len(tmp) < 2, f"Mutiple files with same name `{item}`. Couldn't resolve ambiguity."
            assert len(tmp) > 0, FileNotFoundError(f"FileNotFoundError `{item}`.")
            parent_id = tmp[0]['id']
        return parent_id
    def update(self, remote_path=None, fid=None, local_path=None):
        fid = fid or self.get_id_from_path(remote_path)
        print(f"UPDATING contents of file `{fid}`")
        return self.service.files().update(fileId=fid, media_body=MediaFileUpload(local_path)).execute()

    @staticmethod
    def get_link_from_id(fid, folder=False):
        return tb.P(rf'https://drive.google.com/file/d/{fid}') if not folder else tb.P(rf'https://drive.google.com/drive/u/1/folders/{fid}')
    @staticmethod
    def get_id_from_link(link): return str(link).split(r"https://drive.google.com/file/d/")[1]

    def download(self, fpath=None, fid=None, furl=None, local_dir=None, rel2home=False, decrypt=False, unzip=False, key=None, pwd=None):
        assert fpath or fid or furl, "Either a file name or a file id must be provided."
        if furl is not None: fid = self.get_id_from_link(furl)
        if fpath is not None:
            if rel2home:
                fpath = tb.P(fpath).expanduser().absolute()
                local_dir = fpath.parent
                fpath = tb.get_env().myhome / fpath.rel2home()
            if unzip: fpath = (fpath + ".zip") if ".zip" not in str(fpath) else fpath
            if decrypt: fpath = (fpath + ".enc") if ".enc" not in str(fpath) else fpath
            fid = self.get_id_from_path(fpath)
        else: fpath = self.get_path_from_id(fid)
        local_dir = tb.P(local_dir or f'~/Downloads').expanduser().create()
        fields = self.get_fields_from_id(fid, fields="name, mimeType")
        if fields['mimeType'] == 'application/vnd.google-apps.folder':
            return tb.L(self.service.files().list(q=f"'{fid}' in parents").execute()['files']).apply(lambda x: self.download(fid=x['id'], local_dir=local_dir.joinpath(fields['name'])))
        downloader = MediaIoBaseDownload(fh := io.BytesIO(), self.service.files().get_media(fileId=fid))
        while True:
            status, done = downloader.next_chunk()  # fh: file lives in the RAM now.
            print(f"DOWNLOADING from GDrive {fpath} ==> {local_dir}. {int(status.progress() * 100)}%.")
            if done: break
        fh.seek(0);
        path = local_dir.joinpath(fields['name']).write_bytes(fh.read())
        process_retrieved_file(path, decrypt=decrypt, unzip=unzip, key=key, pwd=pwd)
        return path

    def upload(self, local_path, remote_dir="", overwrite=True, share=False, rel2home=False, zip_first=False, encrypt_first=False, key=None, pwd=None):
        local_path = process_sent_file(file=local_path, zip_first=zip_first, encrypt_first=encrypt_first, key=key, pwd=pwd)
        if rel2home: remote_dir = tb.get_env().myhome.joinpath(local_path.rel2home().parent)
        else: remote_dir = tb.P(remote_dir)
        try: self.get_id_from_path(remote_dir)
        except AssertionError as ae:
            if "FileNotFoundError" in str(ae): self.create_folder(remote_dir)
            else: raise NotImplementedError(f"{ae}")
        print(f"UPLOADING {repr(local_path)} to {repr(remote_dir)} ... ", end="")
        file = {'id': None}
        if local_path.is_dir():
            self.create_folder(path=remote_dir.joinpath(local_path.name))
            for item in local_path.listdir(): self.upload(local_path.joinpath(item), remote_dir=remote_dir.joinpath(local_path.name), zip_first=zip_first, encrypt_first=encrypt_first, key=key, pwd=pwd)
        else:  # upload a file.
            try:
                existing_fid = self.get_id_from_path(remote_dir.joinpath(local_path.name))
                if overwrite: file = self.update(local_path=local_path, fid=existing_fid)
                else: raise FileExistsError(f"File `{local_path.name}` already exists in `{remote_dir}`.")
            except AssertionError as ae:
                if "FileNotFoundError" in str(ae): pass  # good, it's a new file.
                elif "Couldn't resolve ambiguity." in str(ae): raise NotImplementedError("Please manually delete files with same name. " + str(ae))
                file_metadata = {'name': local_path.name, 'parents': [self.get_id_from_path(remote_dir)], 'role': 'reader', 'type': 'anyone'}
                file = self.service.files().create(body=file_metadata, media_body=MediaFileUpload(local_path.str), fields='id').execute()
                print(f"file id: {file.get('id')}")
        if zip_first or encrypt_first: tb.P(local_path).delete(sure=True)
        return {"remote_path": remote_dir.joinpath(local_path.name), 'local_path': local_path, 'fid': file['id']}

    def upload_and_share(self, local_path, overwrite=True, role='writer', zip_first=False, encrypt_first=False, key=None, pwd=None, rel2home=False):
        # https://developers.google.com/drive/api/guides/manage-sharing
        res = self.upload(local_path=local_path, remote_dir="myshare", overwrite=overwrite, share=True, zip_first=zip_first, encrypt_first=encrypt_first, key=key, pwd=pwd, rel2home=rel2home)
        self.service.permissions().create(fileId=res["fid"], body=dict(type='anyone', role=role)).execute()
        res['url'] = tb.P(rf'https://drive.google.com/file/d/{res["fid"]}')
        return res

    def create_folder(self, path="") -> str:
        path = tb.P(path)
        parent_id = 'root'
        for part in path.parts:
            try:
                search_res = self.service.files().list(q=f"'{parent_id}' in parents and name = '{part}' and mimeType = 'application/vnd.google-apps.folder'").execute()['files']
                parent_id = search_res[0]['id']
            except Exception as e:  # not found ==> create it
                file_metadata = {
                    'name': part,
                    'mimeType': 'application/vnd.google-apps.folder',
                    'parents': [parent_id]}
                file = self.service.files().create(body=file_metadata, fields='id').execute()
                print(f"CREATED FOLDER `{part}` in `{path}`. ID = {file.get('id')}")
                parent_id = file.get('id')
        return parent_id


"""
Tip: folders and files on G drive have unique ID that is not the name. Thus, there can be multiple files with same name. 
This can cause seemingly confusing behaviour if the user continues to rely on names rather than IDs. 
Folders in trash bin are particularly confusing as the trash bin is nothing but another folder and thus,
user can still access it and populate it, etc. This action taking place in rubbish bin might not be noticed when inspecting Gdrive via web.
"""


if __name__ == '__main__':
    # pass
    api = GDriveAPI()





from crocodile.core import List, timestamp, Save, install_n_import, validate_name
from crocodile.file_management import P
from crocodile.meta import Terminal

import matplotlib.pyplot as plt
from crocodile.msc.odds import Cycle
from matplotlib import widgets
import matplotlib.colors as mcolors
from matplotlib import animation
import enum
import subprocess
import platform
import pandas as pd
import numpy as np


"""TODO: add implementation https://github.com/gustavovelascoh/plot_update
"""


class FigurePolicy(enum.Enum):
    close_create_new = 'Close the previous figure that has the same figname and create a new fresh one'
    add_new = 'Create a new figure with same path but with added suffix'
    same = 'Grab the figure of the same path'


def assert_requirements():
    try: subprocess.check_output(['where.exe' if platform.system() == 'Windows' else 'which', 'magick'])
    except (FileNotFoundError, subprocess.CalledProcessError):
        # P(r"https://www.gyan.dev/ffmpeg/builds/ffmpeg-git-full.7z").download().unzip().search()[0].rename("ffmpeg").move(r"C://")  # P("C:\\ffmpeg\\bin")  # add to PATH
        print("Installinng image magick")
        if __import__("platform").system() == "Windows":
            Terminal().run("winget install ImageMagick.ImageMagick", shell="powershell")  # gives ffmpeg as well
            print("You might need to restart your machine before PATH change impact takes place.")
        else: raise NotImplementedError


class FigureSave:
    class GenericSave:
        stream = ['clear', 'accumulate', 'update'][0]
        def __init__(self, save_dir=None, save_name=None, watch_figs: list or None = None, max_calls=2000, delay=100, **kwargs):
            """How to control what to be saved: you can either pass the figures to be tracked at init time, pass them dynamically at add time, or, add method will capture every figure and axis"""
            self.watch_figs = watch_figs if watch_figs is None else ([plt.figure(num=afig) for afig in watch_figs] if type(watch_figs[0]) is str else watch_figs)
            self.save_name, self.save_dir = timestamp(name=save_name), save_dir or P.tmpdir(prefix="tmp_fig_save")
            self.kwargs, self.counter, self.delay, self.max = kwargs, 0, delay, max_calls
        def add(self, fignames=None, names=None, **kwargs):  # generic method used at runtime, never changed.
            print(f"Saver added frame number {self.counter}", end='\r')
            self.counter += 1; plt.pause(self.delay * 0.001); print('Turning off IO') if self.counter > self.max else None; plt.ioff()
            self.watch_figs = [plt.figure(figname) for figname in fignames] if fignames else ([plt.figure(k) for k in plt.get_figlabels()] if self.watch_figs is None else self.watch_figs)  # path sent explicitly, # None exist ==> add all else # they exist already.
            if names is None: names = [timestamp(name=a_figure.get_label()) for a_figure in self.watch_figs]  # individual save path, useful for PNG.
            for afig, aname in zip(self.watch_figs, names): self._save(afig, aname, **kwargs)
        def _save(self, *args, **kwargs): pass  # called by generic method `add`, to be implemented when subclassing.
    class Null(GenericSave):  # serves as a filler.
        def __init__(self, *args, **kwargs): super().__init__(*args, **kwargs); self.fname = self.save_dir
        def finish(self): print(f"Nothing saved by {self}"); return self.fname
    class PDF(GenericSave):  # For pdf, you just need any stream [update, clear, accumalate], preferabbly the fastest."""
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            from matplotlib.backends.backend_pdf import PdfPages
            self.fname = self.save_dir.joinpath(self.save_name + ('.pdf' if '.pdf' not in str(self.save_name) else '')); self.pp = PdfPages(self.fname)
        def _save(self, a_fig, a_name, bbox_inches='tight', pad_inches=0.3, **kwargs): self.pp.savefig(a_fig, bbox_inches=bbox_inches, pad_inches=pad_inches, **kwargs)
        def finish(self): print(f"Saving results ..."); self.pp.close(); print(f"SAVED PDF @", P(self.fname).absolute().as_uri()); return self
    class PNG(GenericSave):
        def __init__(self, *args, **kwargs): super().__init__(*args, **kwargs); self.fname = self.save_dir = self.save_dir.joinpath(self.save_name)
        def _save(self, afigure, aname, dpi=150, **kwargs):  afigure.savefig(self.save_dir.joinpath(validate_name(aname)).create(parents_only=True), bbox_inches='tight', pad_inches=0.3, dpi=dpi, **kwargs)
        def finish(self): print(f"SAVED PNGs @", P(self.fname).absolute().as_uri()); return self
    class GIF(GenericSave):  # NOT RECOMMENDED, used GIFFileBased instead.
        """This class uses ArtistAnimation: works on lines and images list attached to figure axes and Doesn't work on axes, unless you add large number of them. As such, titles are not incorporated etc (limitation).
        Requirements: same axis must persist (If you clear the axis, nothing will be saved), only new objects are drawn inside it. Additionally, the objects drawn must not be removed, or updated, instead they should pile up in axis.
        # usually it is smoother when adding animate=True to plot or imshow commands for GIF purpose
        Works for images only. Add more .imshow to the same axis, and that's it. imshow will conver up previous images. For lines, it will superimpose it and will look ugly.
        The class will automatically detect new lines by their "neo" labels and add them then hide them for the next round.
        """
        def __init__(self, interval=100, **kwargs):
            super().__init__(**kwargs); from collections import defaultdict
            self.container, self.interval, self.fname = defaultdict(lambda: []), interval, None  # determined at finish time.
        def _save(self, afigure, aname, cla=False, **kwargs):
            fig_list, subcontainer = self.container[afigure.get_label()], []
            for item in FigureManager.findobj(afigure, 'neo'): item.set_label('processed'); item.set_visible(False); subcontainer += [item]
            fig_list.append(subcontainer)  # if you want the method coupled with cla being used in main, then it add_line is required for axes.
        def finish(self):
            print("Saving the GIF ....")
            for idx, a_fig in enumerate(self.watch_figs):
                if ims := self.container[a_fig.get_label()]:
                    self.fname = self.save_dir.joinpath(f'{a_fig.get_label()}_{self.save_name}.gif')
                    ani = animation.ArtistAnimation(a_fig, ims, interval=self.interval, blit=True, repeat_delay=1000)
                    # noinspection PyTypeChecker
                    ani.save(self.fname, writer=animation.PillowWriter(fps=4))  # if you don't specify the writer, it goes to ffmpeg by default then try others if that is not available, resulting in behaviours that is not consistent across machines.
                    print(f"SAVED GIF @", P(self.fname).absolute().as_uri())
                else: print(f"Nothing to be saved by GIF writer."); return self.fname
    class GIFFileBased(GenericSave):
        def __init__(self, fps=4, dpi=100, bitrate=1800, _type='GIFFileBased', **kwargs):
            super().__init__(**kwargs); assert_requirements()
            if _type == 'GIFPipeBased': writer, extension = animation.ImageMagickFileWriter, '.gif'  # internally calls: matplotlib._get_executable_info("magick")
            elif _type == "GIFFileBased": writer, extension = animation.ImageMagickWriter, '.gif'
            elif _type == 'MPEGFileBased': writer, extension = animation.FFMpegFileWriter, '.mp4'
            elif _type == 'MPEGPipeBased': writer, extension = animation.FFMpegWriter, '.mp4'
            else: raise ValueError("Unknown writer.")
            self.writer = writer(fps=fps, metadata=dict(artist='Alex Al-Saffar'), bitrate=bitrate)
            self.fname = self.save_dir.joinpath(self.save_name + extension)
            assert self.watch_figs, "No figure was sent during instantiation of saver, therefore the writer cannot be setup. Did you mean to use an autosaver?"
            self.writer.setup(fig=self.watch_figs[0], outfile=str(self.fname), dpi=dpi)
        def _save(self, afig, aname, **kwargs): self.writer.grab_frame(**kwargs)
        def finish(self): print('Saving results ...'); self.writer.finish(); print(f"SAVED GIF @", P(self.fname).absolute().as_uri()); return self
    class GIFPipeBased(GIFFileBased):
        def __init__(self, *args, **kwargs): super().__init__(*args, _type=self.__class__.__name__, **kwargs)
    class MPEGFileBased(GIFFileBased):
        def __init__(self, *args, **kwargs): super().__init__(*args, _type=self.__class__.__name__, **kwargs)
    class MPEGPipeBased(GIFFileBased):
        def __init__(self, *args, **kwargs): super().__init__(*args, _type=self.__class__.__name__, **kwargs)
    class GenericAuto(GenericSave):
        """Parses the data internally, hence requires artist with animate method implemetend. Artist needs to have .fig attribute."""
        save_type = 'auto'
        def __init__(self, plotter_class, data, names_list=None, **kwargs):
            super().__init__(**kwargs); self.saver, self.plotter = None, None; assert_requirements()
            self.plotter_class, self.data, self.names_list, self.kwargs = plotter_class, data, names_list, kwargs
        def animate(self):
            self.plotter = self.plotter_class(**self.kwargs); plt.pause(0.5)  # give time for figures to show up before updating them
            for idx, datum in __import__("tqdm").tqdm(enumerate(self.data)): self.plotter.animate(datum); self.saver.add(names=[self.names_list[idx] if self.names_list is not None else str(idx)])
            self.saver.finish()
    class GIFAuto(GenericAuto):
        def __init__(self, plotter_class, data, interval=500, extension='gif', fps=4, metadata=None, **kwargs):
            super().__init__(plotter_class, data, **kwargs)
            writer = animation.PillowWriter(fps=fps) if extension == '.mp4' else animation.FFMpegWriter(fps=fps, metadata=metadata, bitrate=2500)
            self.plotter = self.plotter_class(**kwargs); plt.pause(self.delay * 0.001)  # give time for figures to show up before updating them
            # noinspection PyTypeChecker
            self.ani = animation.FuncAnimation(fig=self.plotter.fig, func=self.plotter.animate, frames=(i for i in zip(*self.data)), interval=interval, repeat_delay=1500, fargs=None, cache_frame_data=True, save_count=10000)
            self.fname = self.save_dir.joinpath(self.save_name + f".{extension}")
            self.ani.save(filename=self.fname, writer=writer); print(f"SAVED GIF @ ", P(self.fname).absolute().as_uri())
    class PDFAuto(GenericAuto):
        def __init__(self, **kwargs): super().__init__(**kwargs); self.saver = FigureSave.PDF(**kwargs); self.animate()
    class PNGAuto(GenericAuto):
        def __init__(self, **kwargs): super().__init__(**kwargs); self.saver = FigureSave.PNG(**kwargs); self.save_dir = self.saver.save_dir; self.animate(); self.fname = self.saver.fname
    class NullAuto(GenericAuto):
        def __init__(self, **kwargs): super().__init__(**kwargs); self.saver = FigureSave.Null(**kwargs); self.fname = self.saver.fname; self.animate()
    class GIFFileBasedAuto(GenericAuto):
        def __init__(self, plotter_class, data, fps=4, dpi=150, bitrate=2500, _type='GIFFileBasedAuto', **kwargs):
            super().__init__(**kwargs)
            if _type == 'GIFPipeBasedAuto': writer = animation.ImageMagickFileWriter; extension = '.gif'
            elif _type == 'MPEGFileBasedAuto': writer = animation.FFMpegFileWriter; extension = '.mp4'
            elif _type == 'MPEGPipeBasedAuto': writer = animation.FFMpegWriter; extension = '.mp4'
            else: raise ValueError("Unknown writer.")
            self.saver = writer(fps=fps, metadata=dict(artist='Alex Al-Saffar'), bitrate=bitrate)
            self.fname = self.save_dir.joinpath(self.save_name + extension)
            self.data = lambda: (i for i in zip(*data)); self.plotter = plotter_class(*[piece[0] for piece in data], **kwargs); plt.pause(0.5); from tqdm import tqdm
            with self.saver.saving(fig=self.plotter.fig, outfile=self.fname, dpi=dpi):
                for datum in tqdm(self.data()): self.plotter.animate(datum); self.saver.grab_frame(); plt.pause(self.delay * 0.001)
            print(f"SAVED GIF successfully @ {self.fname}")
    class GIFPipeBasedAuto(GIFFileBasedAuto):
        def __init__(self, *args, **kwargs): super().__init__(*args, _type=self.__class__.__name__, **kwargs)
    class MPEGFileBasedAuto(GIFFileBasedAuto):
        def __init__(self, *args, **kwargs): super().__init__(*args, _type=self.__class__.__name__, **kwargs)
    class MPEGPipeBasedAuto(GIFFileBasedAuto):
        def __init__(self, *args, **kwargs): super().__init__(*args, _type=self.__class__.__name__, **kwargs)


class FigureManager:  # use as base class for Artist & Viewers to give it free animation powers.
    def __init__(self, info_loc=None, figpolicy=FigurePolicy.same):
        self.figpolicy = figpolicy
        self.fig = self.ax = self.event = None
        self.cmaps, self.colors, self.mcolors, self.facecolor = Cycle(plt.colormaps()), Cycle(plt.rcParams['axes.prop_cycle'].by_key()['color']), list(mcolors.CSS4_COLORS.keys()), Cycle(list(mcolors.CSS4_COLORS.values()))
        self.cmaps.set_value('viridis')
        self.idx_cycle = Cycle([]); self.pause = None  # animation
        self.help_menu = {'_-=+[{]}\\': {'help': "Adjust Vmin Vmax. Shift + key applies change to all axes \\ toggles auto-brightness ", 'func': self.adjust_brightness},
                          "/": {'help': 'Show/Hide info text', 'func': self.text_info},
                          "h": {'help': 'Show/Hide help menu', 'func': self.show_help},
                          "tyTY": {'help': 'Change color map', 'func': self.change_cmap},
                          '<>': {'help': 'Change figure face color', 'func': self.change_facecolor},
                          "v": {'help': 'Show/Hide pixel values (Toggle)', 'func': self.show_pix_val},
                          'P': {'help': 'Pause/Proceed (Toggle)', 'func': self.pause_func},
                          'r': {'help': 'Replay', 'func': self.replay},
                          '1': {'help': 'Previous Image', 'func': self.previous}, '2': {'help': 'Next Image', 'func': self.next},
                          'S': {'help': 'Save Object', 'func': self.save},
                          'c': {'help': 'Show/Hide cursor', 'func': self.show_cursor},
                          'aA': {'help': 'Show/Hide ticks and their labels', 'func': self.show_ticks},
                          'alt+a': {'help': 'Show/Hide annotations', 'func': self.toggle_annotate}}  # IMPORTANT: add the 'alt/ctrl+key' versions of key after the key in the dictionary above, not before, otherwise the naked key version will statisfy the condition `is key in this`? in the parser.
        self.auto_brightness, self.pix_vals = False, False; self.boundaries_flag, self.annot_flag = True, False
        self.info_loc = [0.8, 0.01] if info_loc is None else info_loc
        self.message, self.message_obj, self.cursor = '', None, None
    def show_help(self, event):
        default_plt = {"q ": {'help': "Quit Figure."},
                       "Ll": {'help': "change x/y scale to log and back to linear (toggle)"},
                       "Gg": {'help': "Turn on and off x and y grid respectively."},
                       "s ": {'help': "Save Figure"},
                       "f ": {'help': "Toggle Full screen"},
                       "p ": {'help': "Select / Deselect Pan"}}
        if "Keyboard shortcuts" in plt.get_figlabels(): plt.close("Keyboard shortcuts"); _ = event  # toggle
        else:
            fig = plt.figure(num="Keyboard shortcuts")
            for i, key in enumerate(self.help_menu.keys()): fig.text(0.1, 1 - 0.05 * (i + 1), f"{key:30s} {self.help_menu[key]['help']}")
            print(pd.DataFrame([[val['help'], key] for key, val in self.help_menu.items()], columns=['Action', 'Key']), "\nDefault plt Keys:\n")
            print(pd.DataFrame([[val['help'], key] for key, val in default_plt.items()], columns=['Action', 'Key']))
    # =============== EVENT METHODS ====================================
    def animate(self): pass  # a method of the artist child class that is inheriting from this class to define behaviour when user press next or previous buttons.
    def connect(self): self.fig.canvas.mpl_connect('key_press_event', self.process_key); return self
    def process_key(self, event):
        self.event = event  # useful for debugging.
        for key in self.help_menu.keys():
            if event.key in key: self.help_menu[key]['func'](event); self.update_info_text(self.message); break
        if event.key != 'q': event.canvas.figure.canvas.draw()  # for smooth quit without throwing errors  # don't update if you want to quit.
    def toggle_annotate(self, event):
        self.annot_flag = not self.annot_flag
        if event.inaxes and event.inaxes.images: event.inaxes.images[0].set_picker(True); self.message = f"Annotation flag is toggled to {self.annot_flag}"
    def annotate(self, event, axis=None, data=None):
        self.event = event; e = event.mouseevent; ax = e.inaxes if axis is None else axis
        if not ax: return None
        if not hasattr(ax, 'annot_obj'): ax.annot_obj = ax.annotate("", xy=(0, 0), xytext=(-30, 30), textcoords="offset points", arrowprops=dict(arrowstyle="->", color="w", connectionstyle="arc3"),
                                                                    va="bottom", ha="left", fontsize=10, bbox=dict(boxstyle="round", fc="w"), )
        else: ax.annot_obj.set_visible(self.annot_flag)
        x, y = int(np.round(e.xdata)), int(np.round(e.ydata))
        z = e.inaxes.images[0].get_array()[y, x] if data is None else data[y, x]
        ax.annot_obj.set_text(f'x:{x}\ny:{y}\nvalue:{z:.3f}'); ax.annot_obj.xy = (x, y); self.fig.canvas.draw_idle()
    def save(self, event): _ = event; Save.pickle(path=P.tmpfile(name="figure_manager"), obj=self)
    def replay(self, event): _ = event; self.pause = False; self.idx_cycle.set_index(0); self.message = 'Replaying'; self.animate()
    def pause_func(self, event): _ = event; self.pause = not self.pause; self.message = f'Pause flag is set to {self.pause}'; self.animate()
    def previous(self, event): _ = event; self.idx_cycle.previous(); self.message = f'Previous {self.idx_cycle}'; self.animate()
    def next(self, event): _ = event; self.idx_cycle.next(); self.message = f'Next {self.idx_cycle}'; self.animate()
    def text_info(self, event): _ = event; self.message = ''
    def change_facecolor(self, event): self.fig.set_facecolor(self.facecolor.next() if event.key == '>' else self.facecolor.previous()); self.message = f"Figure facecolor was set to {self.mcolors[self.facecolor.get_index()]}"
    def adjust_brightness(self, event):
        ax, message = event.inaxes, "None"
        if ax is None or not ax.images: return None
        if event.key == '\\':
            self.auto_brightness = not self.auto_brightness; message = f"Auto-brightness flag is toggled to {self.auto_brightness}"
            if self.auto_brightness: im = self.ax.images[0]; im.norm.autoscale(im.get_array())  # changes to all ims take place in animate as in ImShow and Nifti methods animate.
        vmin, vmax = ax.images[0].get_clim()
        if event.key in '-_': message = 'increase vmin'; vmin += 1
        elif event.key in '[{': message = 'decrease vmin'; vmin -= 1
        elif event.key in '=+': message = 'increase vmax'; vmax += 1
        elif event.key in ']}': message = 'decrease vmax'; vmax -= 1
        self.message = message + '  ' + str(round(vmin, 1)) + '  ' + str(round(vmax, 1))
        if event.key in '_+}{': [ax.images[0].set_clim((vmin, vmax)) for ax in self.fig.axes if ax.images]
        else: ax.images[0].set_clim((vmin, vmax)) if ax.images else None
    def change_cmap(self, event):
        if ax := event.inaxes is not None:
            cmap = self.cmaps.next() if event.key in 'tT' else self.cmaps.previous()
            [[im.set_cmap(cmap) for im in ax.images] for ax in self.fig.axe] if event.key in 'TY'else [im.set_cmap(cmap) for im in ax.images]
            self.message = f"Color map changed to {ax.images[0].cmap.name}"
    def show_pix_val(self, event):
        if (ax := event.inaxes) is not None:
            self.pix_vals = not self.pix_vals; self.message = f"Pixel values flag set to {self.pix_vals}"
            if self.pix_vals: self.show_pixels_values(ax)
            else:
                while len(ax.texts) > 0: [text.remove() for text in ax.texts]
    def show_cursor(self, event):
        if not (ax := event.inaxes): return None  # don't do this if c was pressed outside an axis.
        if hasattr(ax, 'cursor_'):  # is this the first time?
            if ax.cursor_ is None: ax.cursor_ = widgets.Cursor(ax=ax, vertOn=True, horizOn=True, color='red', lw=1.0)
            else: ax.cursor_ = None  # toggle the cursor.
            self.message = f'Cursor flag set to {bool(ax.cursor_)}'
        else: ax.cursor_ = None; self.show_cursor(event)  # first call
    def show_ticks(self, event):
        self.boundaries_flag = not self.boundaries_flag
        if event.key == 'a' and (axis := event.inaxes): self.toggle_ticks(axis); self.message = f"Boundaries flag set to {self.boundaries_flag} in {axis}"
        else: [self.toggle_ticks(ax) for ax in self.ax]
    # ====================== class methods ===============================
    def get_fig(self, figname='', suffix=None, **kwargs): return FigureManager.get_fig_static(self.figpolicy, figname, suffix, **kwargs)
    def update_info_text(self, message): self.message_obj.remove() if self.message_obj else None; self.message_obj = self.fig.text(*self.info_loc, message, fontsize=8)
    def maximize_fig(self): _ = self; plt.get_current_fig_manager().full_screen_toggle()  # TODO not working appropriately ImShow.test() # The command required is backend-dependent and also OS dependent. Doesn't work if figure is not shown yet.
    def clear_axes(self): [ax.cla() for ax in self.ax]
    def transperent_fig(self): self.fig.canvas.manager.window.attributes("-transparentcolor", "white")
    def close(self): plt.close(self.fig)
    # ====================== axis helpers ========================
    @staticmethod
    def grid(ax, factor=5, x_or_y='both', color='gray', alpha1=0.5, alpha2=0.25):
        if type(ax) in {list, List, np.ndarray}: [FigureManager.grid(an_ax, factor=factor, x_or_y=x_or_y, color=color, alpha1=alpha1, alpha2=alpha2) for an_ax in ax]  # Turning on major grid for both axes.
        ax.grid(which='major', axis='x', color='gray', linewidth=0.5, alpha=alpha1); ax.grid(which='major', axis='y', color='gray', linewidth=0.5, alpha=alpha1)
        if x_or_y in {'both', 'x'}: xt = ax.get_xticks(); ax.xaxis.set_minor_locator(plt.MultipleLocator((xt[1] - xt[0]) / factor)); ax.grid(which='minor', axis='x', color=color, linewidth=0.5, alpha=alpha2)
        if x_or_y in {'both', 'y'}: yt = ax.get_yticks(); ax.yaxis.set_minor_locator(plt.MultipleLocator((yt[1] - yt[0]) / factor)); ax.grid(which='minor', axis='y', color=color, linewidth=0.5, alpha=alpha2)
    @staticmethod
    def set_ax_size(ax, w, h, units='inches'):
        l, r, t, b, _ = ax.figure.subplotpars.left, ax.figure.subplotpars.right, ax.figure.subplotpars.top, ax.figure.subplotpars.bottom, units
        ax.figure.set_size_inches(float(w) / (r - l), float(h) / (t - b))
    @staticmethod
    def get_ax_size(ax, units="inches"):
        w, h = ax.figure.get_size_inches() if units == "inches" else ax.figure.get_size_pixels()
        width, height = ax.figure.subplotpars.right - ax.figure.subplotpars.left, ax.figure.subplotpars.top - ax.figure.subplotpars.bottom
        return w * width, h * height
    @staticmethod
    def toggle_ticks(an_ax, state=None): [line.set_visible(not line.get_visible() if state is None else state) for line in an_ax.get_yticklines() + an_ax.get_xticklines() + an_ax.get_xticklabels() + an_ax.get_yticklabels()]
    @staticmethod
    def set_ax_to_real_life_size(ax, inch_per_unit=1 / 25.4): FigureManager.set_ax_size(ax, (ax.get_xlim()[1] - ax.get_xlim()[0]) * inch_per_unit, (ax.get_ylim()[1] - ax.get_ylim()[0]) * inch_per_unit)
    @staticmethod
    def show_pixels_values(ax):
        xmin, xmax = ax.get_xlim(); ymin, ymax = ax.get_ylim()
        if ymin > ymax: ymin, ymax = ymax, ymin  # default imshow settings
        [ax.text(i, j, np.round(label).__int__(), ha='center', va='center', size=8) for (j, i), label in np.ndenumerate(ax.images[0].get_array()) if (xmin < i < xmax) and (ymin < j < ymax)]
    # ============================ matplotlib setup ==============================
    @staticmethod
    def get_nrows_ncols(num_plots, nrows=None, ncols=None):
        if not nrows and not ncols:
            nrows, ncols = int(np.floor(np.sqrt(num_plots))), int(np.ceil(np.sqrt(num_plots)))
            while nrows * ncols < num_plots: ncols += 1
        elif not ncols and nrows: ncols = int(np.ceil(num_plots / nrows))
        elif not nrows and ncols: nrows = int(np.ceil(num_plots / ncols))
        return nrows, ncols
    @staticmethod
    def findobj(figname, obj_name): return (plt.figure(num=figname) if type(figname) is str else figname).findobj(lambda x: x.get_label() == obj_name)
    @staticmethod
    def get_fig_static(figpolicy, figname='', suffix=None, **kwargs):
        exist = True if figname in plt.get_figlabels() else False
        if figpolicy is FigurePolicy.same: return plt.figure(num=figname, **kwargs)
        elif figpolicy is FigurePolicy.add_new: return plt.figure(num=(timestamp(name=figname) if suffix is None else figname + suffix) if exist else figname, **kwargs)
        elif figpolicy is FigurePolicy.close_create_new: plt.close(figname) if exist else None; return plt.figure(num=figname, **kwargs)
    @staticmethod
    def try_figure_size():
        fig, ax = plt.subplots()
        y = np.sin(x := np.arange(0, 100, 0.01)) * 100
        ax.plot(x, y); ax.axis("square"); ax.set_xlim(0, 100); ax.set_ylim(-100, 100)
        FigureManager.set_ax_to_real_life_size(ax); fig.savefig(P.tmp() / "trial.png", dpi=250)
    @staticmethod
    def write(txt, name="text", size=8, **kwargs):
        FigureManager.maximize_fig(fig := plt.figure(figsize=(11.69, 8.27), num=name))
        fig.clf(); fig.text(0.5, 0.5, txt, transform=fig.transFigure, size=size, ha="center", va='center', **kwargs); return fig
    @staticmethod
    def activate_latex(size=20):
        plt.rc('xtick', labelsize=size); plt.rc('ytick', labelsize=size)
        plt.rc('axes', titlesize=size); plt.rc('legend', fontsize=size / 1.5)  # rc('text', usetex=True)
        plt.rcParams['text.usetex'] = True; plt.rc('font', **{'family': 'serif', 'serif': ['Computer Modern']})
    @staticmethod
    def set_linestyles_and_markers_and_colors(test=False):
        from cycler import cycler; from matplotlib import lines
        markers = list(lines.lineMarkers.keys())[:-4]  # ignore the None
        linestyles = (list(lines.lineStyles.keys())[:-3] * 10)[:len(markers)]
        colors = (plt.rcParams['axes.prop_cycle'].by_key()['color'] * 10)[:len(markers)]
        default_cycler = (cycler(linestyle=linestyles) + cycler(marker=markers) + cycler(color=colors))
        plt.rc('axes', prop_cycle=default_cycler)
        if test: [plt.plot(aq + idx * 2) for idx, aq in enumerate(np.random.randn(10, 10))]


class VisibilityViewer(FigureManager):  # This is used for browsing purpose, as opposed to saving.
    artist = ['internal', 'external'][1]  # How is the data visualized? You need artist. The artist can either be internal, as in ImShow or passed externally (especially non image data)
    parser = ['internal', 'external'][1]  # Data parsing: internal for loop to go through all the dataset passed. # Allows manual control over parsing. external for loop. It should have add method. # Manual control only takes place after the external loop is over. #TODO parallelize this.
    stream = ['clear', 'accumulate', 'update'][1]  # Streaming (Refresh mechanism): * Clear the axis. (slowest, but easy on memory) * accumulate, using visibility to hide previous axes. (Fastest but memory intensive)  * The artist has an update method. (best)
    """ This class works on hiding axes shown on a plot, so that a new plot can be drawn. Once the entire loop is finished, you can browse through the plots with the keyboard Animation linked to `animate`"""
    def __init__(self, fig):  # Downside: slow if number of axes, lines, texts, etc. are too large. In that case, it is better to avoid this viewer and plot on the fly during animation.
        super().__init__(); self.objs_repo = []  # list of lists of axes and texts.
        self.fig = fig; self.connect(); self.idx_cycle = Cycle([])
    def add(self, objs): self.idx_cycle.expand(); self.objs_repo.append(objs); [obj.set_visible(False) for obj in (self.objs_repo[-2] if len(self.objs_repo) > 1 else [])]; print(f"VViewer added plot number {self.idx_cycle.get_index()}", end='\r')
    def animate(self): [ax.set_visible(False) for ax in self.objs_repo[self.idx_cycle.prev_index]]; [ax.set_visible(True) for ax in self.objs_repo[self.idx_cycle.get_index()]]; self.fig.canvas.draw()


class Artist(FigureManager):  # This object knows how to draw a figure from curve-type data.
    def __init__(self, ax=None, figname='Graph', title='', label='curve', style='seaborn', figpolicy=FigurePolicy.add_new, figsize=(14, 8)):
        super().__init__(figpolicy=figpolicy)
        self.style, self.title = style, title; self.line = self.cursor = self.check_b = None
        if ax is None:  # create a figure
            with plt.style.context(style=self.style): self.fig = self.get_fig(figname, figsize=figsize); self.ax = self.fig.subplots()
        else: self.ax = ax; self.fig = ax.figure  # use the passed axis
        self.visibility_ax, self.txt, self.label = [0.01, 0.05, 0.2, 0.15], [], label
    def plot(self, *args, legends=None, title=None, **kwargs): self.line = self.ax.plot(*args, **kwargs); self.ax.legend(legends or []); self.ax.set_title(title) if title is not None else None; self.ax.grid('on')
    def plot_dict(self, adict, title='', xlabel='', ylabel=''): [self.plot(val, label=key) for key, val in adict.items()]; self.ax.legend(); self.ax.set_title(title); self.ax.set_xlabel(xlabel); self.ax.set_ylabel(ylabel); return self
    def plot_twin(self, c1, c2, x=None, l1='', l2='', ax=None): ax = ax or self.ax; twin_ax = ax.twinx(); line1 = ax.plot(x or range(len(c1)), c1, color="blue", label=l1)[0]; line2 = twin_ax.plot(c2, color="red", label=l2)[0]; twin_ax.legend([line1, line2], [l1, l2]); ax.set_ylabel(l1); twin_ax.set_ylabel(l2); plt.show()
    def suptitle(self, title): self.txt = [self.fig.text(0.5, 0.98, title, ha='center', va='center', size=9)]
    def clear(self): self.fig.clf()  # objects that use this class will demand that the artist expose this method, in addition to .plot()
    def axes(self): return [self.ax]  # objects that use this class will demand that the artist expose this method, in addition to .plot()
    def visibility(self):
        from matplotlib.widgets import CheckButtons; self.fig.subplots_adjust(left=0.3); self.visibility_ax[-1] = 0.05 * len(self.ax.lines)
        rax = self.fig.add_axes(self.visibility_ax)
        labels, visibility = [str(line.get_label()) for line in self.ax.lines], [line.get_visible() for line in self.ax.lines]
        self.check_b = CheckButtons(rax, labels, visibility)
        def func(label): index = labels.index(label); self.ax.lines[index].set_visible(not self.ax.lines[index].get_visible()); self.fig.canvas.draw()
        self.check_b.on_clicked(func)
    @staticmethod
    def styler(plot_gen):
        for astyle in plt.style.available:
            with plt.style.context(style=astyle): plot_gen(); plt.title(astyle); plt.pause(1); plt.cla()


class VisibilityViewerAuto(VisibilityViewer):
    artist = ['internal', 'external'][1]
    parser = ['internal', 'external'][0]
    stream = ['clear', 'accumulate', 'update'][0:2]
    def __init__(self, data=None, artist=None, stream='clear', save_type=FigureSave.Null, save_dir=None, save_name=None, delay=1,
                 titles=None, legends=None, x_labels=None, pause=True, **kwargs):
        """data: tensor of form  NumInputsForAnimation x ArgsPerPlot (to be unstarred) x Input (possible points x signals)
        stream: ensure that behaviour of artist is consistent with stream. When `cccumulate`, artist should create new axes whenever plot is called."""
        self.data = data; self.artist = artist or Artist()
        self.legends = [f"Plot {i}" for i in range(len(self.data))] if legends is None else legends; self.titles = titles if titles is not None else np.arange(len(self.data))
        super().__init__(fig=self.artist.fig); _ = kwargs
        self.saver = save_type(watch_figs=[self.artist.fig], save_dir=save_dir, save_name=save_name, delay=delay, fps=1000 / delay)
        self.index_max, self.pause, self.stream, self.lables = len(self.data), pause, stream, x_labels
    test = staticmethod(lambda: VisibilityViewerAuto(data=np.random.randn(10, 1, 10, 3)))
    def animate(self):
        for i in range(0, self.index_max):
            self.artist.plot(*self.data[i], title=self.titles[i], legends=self.legends)  # replot the new data point on a new axis.
            super().add(self.artist.axes()) if self.stream == 'accumulate' else self.artist.clear()
            self.saver.add()
            if self.pause: break
            else: self.idx_cycle.set_index(i)
        return self.saver.finish()  # arrived at last image and not in manual mode


class ImShow(FigureManager):
    artist = ['internal', 'external'][0]
    parser = ['internal', 'external'][0]
    stream = ['clear', 'accumulate', 'update'][2]
    def __init__(self, img_tensor, sup_titles=None, sub_labels=None, save_type=FigureSave.Null, save_name=None, save_dir=None, save_kwargs=None,
                 subplots_adjust=None, gridspec=None, tight=True, info_loc=None, nrows=None, ncols=None, ax=None,
                 figsize=None, figname='im_show', auto_brightness=True, delay=200, pause=False, **kwargs):
        """
        :param img_tensor: size N x M x W x H [x C]  # M used spatially, N for animation.
        :param sup_titles: Titles for frames (N)
        :param sub_labels: M x N. If shape sent is M
        """
        n, m = len(img_tensor), len(img_tensor[0]); self.m, self.n = m, n; super(ImShow, self).__init__(info_loc=info_loc)
        nrows, ncols = self.get_nrows_ncols(m, nrows, ncols)
        self.img_tensor, self.sub_labels, self.sup_titles = img_tensor, sub_labels if sub_labels is not None else [[f"{i}-{j}" for j in range(m)] for i in range(n)], sup_titles if sup_titles is not None else np.arange(n)
        self.pause, self.kwargs, self.delay, self.auto_brightness = pause, kwargs, delay, auto_brightness
        self.fname = self.event = None; self.ims = []  # container for images.
        self.cmaps = Cycle(plt.colormaps()); self.cmaps.set_value('viridis')
        if ax is None:
            self.fig = self.get_fig(figname=figname, figsize=(14, 9) if figsize is None else figsize, facecolor='white'); self.maximize_fig() if figsize is None else None
            if gridspec is not None: gs = self.fig.add_gridspec(gridspec[0]); self.ax = [self.fig.add_subplot(gs[ags[0], ags[1]]) for ags in gs[1:]]
            else: self.ax = self.fig.subplots(nrows=nrows, ncols=ncols)
        else: self.ax = ax; self.fig = ax[0].figure if type(ax) is list else ax.figure
        self.connect(); self.fig.canvas.mpl_connect("pick_event", self.annotate)
        self.fig.tight_layout() if tight else None; self.fig.subplots_adjust(**subplots_adjust) if subplots_adjust is not None else None
        self.saver = save_type(watch_figs=[self.fig], save_dir=save_dir, save_name=save_name, delay=delay, fps=1000 / delay, **({} if save_kwargs is None else save_kwargs))
        self.ax = [self.ax] if nrows == 1 and ncols == 1 else self.ax.ravel()  # make a list out of it or # make a 1D  list out of a 2D array.
        [self.toggle_ticks(an_ax, state=False) for an_ax in self.ax]; self.idx_cycle = Cycle(max_idx=len(self.img_tensor))  # self.animate()
    def animate(self):
        for i in range(self.idx_cycle.get_index(), self.n):
            for j, (an_image, a_label, an_ax) in enumerate(zip(self.img_tensor[i], self.sub_labels[i], self.ax)):  # with zipping, the shortest of the three, will stop the loop.
                if i == 0 and self.ims.__len__() < self.m: self.ims.append(an_ax.imshow(an_image, animated=True, **self.kwargs))
                else: self.ims[j].set_data(an_image)
                if self.auto_brightness: self.ims[j].norm.autoscale(an_image)
                an_ax.set_xlabel(f'{a_label}')
            self.fig.suptitle(self.sup_titles[i], fontsize=8); self.saver.add(names=[self.sup_titles[i]])
            if self.pause: break
            else: self.idx_cycle.set_index(i)
        if self.idx_cycle.get_index() == self.n - 1 and not self.pause: self.fname = self.saver.finish()  # arrived at last image and not in manual mode
    @staticmethod
    def try_cmaps(im, nrows=3, ncols=7, **kwargs): _ = ImShow(*np.array_split([plt.get_cmap(style)(im) for style in plt.colormaps()], nrows * ncols), nrows=nrows, ncols=ncols, sub_labels=np.array_split(plt.colormaps(), nrows * ncols), **kwargs); return [plt.get_cmap(style)(im) for style in plt.colormaps()]
    def annotate(self, event, axis=None, data=None): [super().annotate(event, axis=ax, data=ax.images[0].get_array()) for ax in self.ax]
    from_img_paths = staticmethod(lambda paths, **kwargs: ImShow(List(paths).apply(plt.imread), sub_labels=List(paths).apply(lambda x: P(x).stem), **kwargs))
    from_complex = staticmethod(lambda data, pause=True, **kwargs: ImShow(data.real, data.imag, np.angle(data), abs(data), labels=['Real Part', 'Imaginary Part', 'Angle in Radians', 'Absolute Value'], pause=pause, **kwargs))
    test = staticmethod(lambda: ImShow(np.random.rand(12, 10, 80, 120, 3)))  # https://ai.googleblog.com/2019/08/turbo-improved-rainbow-colormap-for.html # https://gist.github.com/mikhailov-work/ee72ba4191942acecc03fe6da94fc73f
    resize = staticmethod(lambda path, m, n: plt.imsave(path, install_n_import(package="skimage", name="scikit-image").transform.resize(plt.imread(path), (m, n), anti_aliasing=True)))


if __name__ == '__main__':
    pass




import crocodile.toolbox as tb
import platform
import getpass
import os
import sys

P = tb.P
L = tb.L

system = platform.system()  # Linux or Windows
myhome = tb.P(f"myhome/{platform.system().lower()}")

OS = os.getenv("OS")  # Windows_NT
sep = ";" if system == "Windows" else ":"  # PATH separator, this is special for PATH object, not to be confused with P.sep (normal paths), usually / or \
# env = tb.Struct(dict(os.environ)).clean_view
exe = P(sys.executable)

tm = tb.Terminal()

# ============================== Common Paths ============================

LocalAppData = P(tmp) if (tmp := os.getenv("LOCALAPPDATA")) else None  # C:\Users\username\AppData\Local
AppData = P(tmp) if (tmp := os.getenv("APPDATA")) else None  # C:\Users\username\AppData\Roaming
WindowsApps = LocalAppData.joinpath(r"Microsoft\WindowsApps") if AppData else None  # this path is already in PATH. Thus, useful to add symlinks and shortcuts to apps that one would like to be in the PATH.

ProgramData = P(tmp) if (tmp := os.getenv("PROGRAMDATA")) else None  # C:\ProgramData
ProgramFiles = P(tmp) if (tmp := os.getenv("ProgramFiles")) else None  # C:\Program Files
ProgramW6432 = P(tmp) if (tmp := os.getenv("ProgramW6432")) else None  # C:\Program Files
ProgramFilesX86 = P(tmp) if (tmp := os.getenv("ProgramFiles(x86)")) else None  # C:\Program Files (x86)

CommonProgramFiles = P(tmp) if (tmp := os.getenv("CommonProgramFiles")) else None  # C:\Program Files\Common Files
CommonProgramW6432 = P(tmp) if (tmp := os.getenv("CommonProgramW6432")) else None  # C:\Program Files\Common Files
CommonProgramFilesX86 = P(tmp) if (tmp := os.getenv("CommonProgramFiles(x86)")) else None  # C:\Program Files (x86)\Common Files

Tmp = P(tmp) if (tmp := os.getenv("TMP")) else None  # C:\Users\usernrame\AppData\Local\Temp
Temp = Tmp

PATH = Path = L(os.getenv("PATH").split(sep)).apply(P)
PSPath = L(tmp.split(sep)).apply(P) if (tmp := os.getenv("PSModulePath")) else None

HostName          = platform.node()  # e.g. "MY-SURFACE", os.getenv("COMPUTERNAME") only works for windows.
UserName          = getpass.getuser()  # e.g: username, os.getenv("USERNAME") only works for windows.
UserDomain        = os.getenv("USERDOMAIN")  # e.g. HAD OR MY-SURFACE
UserDomainRoaming = P(tmp) if (tmp := os.getenv("USERDOMAIN_ROAMINGPROFILE")) else None  # e.g. SURFACE
LogonServer       = os.getenv("LOGONSERVER")  # e.g. "\\MY-SURFACE"
# UserProfile       = P(tmp) if (tmp := os.getenv("USERPROFILE")) else None  # e.g C:\Users\username
# HomePath          = P(tmp) if (tmp := os.getenv("HOMEPATH")) else None  # e.g. C:\Users\username
Public            = P(tmp) if (tmp := os.getenv("PUBLIC")) else None  # C:\Users\Public

WSL_FROM_WIN = P(r"\\wsl.localhost\Ubuntu-22.04\home")  # P(rf"\\wsl$\Ubuntu\home")  # see localappdata/canonical
WIN_FROM_WSL = P(rf"/mnt/c/Users")


OneDriveConsumer = P(tmp) if (tmp := os.getenv("OneDriveConsumer")) else None
OneDriveCommercial = P(tmp) if (tmp := os.getenv("OneDriveCommercial")) else None
OneDrive = P(tmp) if (tmp := os.getenv("OneDrive")) else None
if system == "Windows":
    tmp1 = LocalAppData.joinpath("Microsoft/OneDrive/OneDrive.exe")
    tmp2 = P(r"C:/Program Files/Microsoft OneDrive/OneDrive.exe")
    OneDriveExe = tmp1 if tmp1.exists() else tmp2
else: OneDriveExe = None


DotFiles = P.home().joinpath("dotfiles")

# ============================== Networking ==============================


def get_addresses():
    netifaces = tb.install_n_import("netifaces")
    subnet_mask = netifaces.ifaddresses(netifaces.gateways()['default'][netifaces.AF_INET][1])[netifaces.AF_INET][0]['netmask']
    default_gateway = netifaces.gateways()['default'][netifaces.AF_INET][0]
    import uuid
    mac = uuid.getnode()
    mac_address = ":".join(("%012X" % mac)[i:i + 2] for i in range(0, 12, 2))
    # elif hex_format: return hex(mac)
    # else: return mac
    import socket
    local_ip_v4 = socket.gethostbyname(socket.gethostname() + ".local")  # without .local, in linux machines, '/etc/hosts' file content, you have an IP address mapping with '127.0.1.1' to your hostname
    return dict(subnet_mask=subnet_mask, mac_address=mac_address, local_ip_v4=local_ip_v4, default_gateway=default_gateway, public_ip_v4=P('https://api.ipify.org').download(memory=True).text)


# ============================== System Variables ==============================


class ShellVar(object):
    @staticmethod
    def set(key, val, run=False, shell="powershell"):
        if system == "Windows":
            res = f"set {key} {val}"
            return res if not run else tm.run(res, shell=shell)
        elif system == "Linux":
            res = f"{key} = {val}"
            return res if not run else tm.run(res, shell="bash")

    @staticmethod
    def get(key, run=False):
        result = f"${key}"  # works in powershell and bash
        return result if run is False else tm.run(result, shell="powershell")
    # in windows cmd `%key%`


class EnvVar:
    @staticmethod
    def set(key, val, temp=False, run=False):
        if system == "Windows":
            if temp is False:
                res = f"setx {key} {val}"  # WARNING: setx limits val to 1024 characters # in case the variable included ";" separated paths, this limit can be exceeded.
                return res if not run else tm.run(res, shell="powershell")
            else: raise NotImplementedError
        elif system == "Linux": return f"export {key} = {val}"  # this is shell command. in csh: `setenv key val`
        else: raise NotImplementedError

    @staticmethod
    def get(key, run=False):
        result = f"${key}"  # works in powershell and bash
        return result if run is False else tm.run(result, shell="powershell")

    # in windows cmd `%key%`
    @staticmethod
    def delete(key, temp=True, scope=["User", "system"][0], run=False):
        if system == "Windows":
            if temp:
                result = fr"Remove-Item Env:\{key}"  # temporary removal (session)
                return result if run is False else tm.run(result, shell="powershell")
            else:
                result = fr'[Environment]::SetEnvironmentVariable("{key}",$null,"{scope}")'
                return result if run is False else tm.run(result, shell="powershell")
        else:
            raise NotImplementedError


class PathVar:
    @staticmethod
    def append_temporarily(dirs, kind="append", run=False):
        dirs_ = []
        for path in dirs:
            path_rel = tb.P(path).collapseuser(strict=False)
            if path_rel.as_posix() in Path or str(path_rel) in Path or path_rel.expanduser().str in Path or path_rel.expanduser().as_posix() in Path: print(f"Path passed `{path}` is already in PATH, skipping the appending.")
            else: dirs_.append(path_rel.as_posix())
        dirs = dirs_
        if len(dirs) == 0: return ""

        if system == "Windows":
            """Source: https://docs.microsoft.com/en-us/powershell/module/microsoft.powershell.core/about/about_environment_variables?view=powershell-7.2"""
            if kind == "append": command = fr'$env:Path += ";{sep.join(dirs)}"'  # Append to the Path variable in the current window:
            elif kind == "prefix": command = fr'$env:Path = "{sep.join(dirs)};" + $env:Path'  # Prefix the Path variable in the current window:
            elif kind == "replace": command = fr'$env:Path = "{sep.join(dirs)}"'  # Replace the Path variable in the current window (use with caution!):
            else: raise KeyError
            return command if run is False else tm.run(command, shell="powershell")
        elif system == "Linux": result = f'export PATH="{sep.join(dirs)}:$PATH"'
        else: raise ValueError
        return result if run is False else tm.run(result, shell="powershell")

    @staticmethod
    def append_permanently(path, scope=["User", "system"][0], run=False):
        if system == "Windows":
            # AVOID THIS AND OPT TO SAVE IT IN $profile.
            tmp_path = tb.P.tmpfile(suffix=".path_backup")
            if tb.P(path) in Path:
                print(f"Path passed `{path}` is already in PATH, skipping the appending.")
                return None
            backup = fr'$env:PATH >> {tmp_path}; '
            command = fr'[Environment]::SetEnvironmentVariable("Path", $env:PATH + ";{path}", "{scope}")'
            result = backup + command
            return result if run is False else tm.run(result, shell="powershell").print()
        else: tb.P.home().joinpath(".bashrc").append_text(f"export PATH='{path}:$PATH'")

    @staticmethod
    def set_permanetly(path, scope=["User", "system"][0], run=False):
        """This is useful if path is manipulated with a text editor or Python string manipulation (not recommended programmatically even if original is backed up) and set the final value.
        On a windows machine, system and user variables are kept separately. env:Path returns the combination of both, starting from system then user.
        To see impact of change, you will need to restart the process from which the shell started. This is probably windows explorer.
        This can be achieved by suspending the process, alternatively you need to logoff and on.
        This is because enviroment variables are inherited from parent process, and so long explorere is not updated, restarting the shell would not help."""
        tmpfile = tb.P.tmpfile(suffix=".path_backup")
        print(f"Saving original path to {tmpfile}")
        backup = fr'$env:PATH >> {tmpfile}; '
        result = backup + fr'[Environment]::SetEnvironmentVariable("Path", "{path}", "{scope}")'
        return result if run is False else tm.run(result, shell="powershell")

    @staticmethod
    def load_fresh_path(run=False):
        if system == "Windows":
            result = fr'[System.Environment]::GetEnvironmentVariable("Path","Machine") + ";" + [System.Environment]::GetEnvironmentVariable("Path","User")'
            return result if run is False else tm.run(result, shell="powershell")


# ============================== Shells =========================================


def get_shell_profiles(shell):
    # following this: https://docs.microsoft.com/en-us/powershell/module/microsoft.powershell.core/about/about_profiles?view=powershell-7.2
    # https://devblogs.microsoft.com/scripting/understanding-the-six-powershell-profiles/
    # Dynmaically obtained:
    return tb.Struct(
        CurrentUserCurrentHost=tm.run("$PROFILE.CurrentUserCurrentHost", shell=shell).op2path(),
        CurrentUserAllHosts=tm.run("$PROFILE.CurrentUserAllHosts", shell=shell).op2path(),
        AllUsersCurrentHost=tm.run("$PROFILE.AllUsersCurrentHost", shell=shell).op2path(),
        AllUsersAllHosts=tm.run("$PROFILE.AllUsersAllHosts", shell=shell).op2path(),
    )


def construct_path(path_list): return tb.L(__import__("pd").unique(path_list)).reduce(lambda x, y: str(x) + sep + str(y))
def get_path_defined_files(string_="*.exe"): res = Path.search(string_).reduce(lambda x, y: x + y); res.print(); return res


if __name__ == '__main__':
    pass





"""Crocodile Shell
"""

import argparse
import os
import random
from rich import pretty, inspect, progress, traceback, print
from rich.text import Text
from rich.console import Console

from crocodile.core import *
from crocodile.file_management import *
from crocodile.meta import *
# import crocodile.environment as env
from crocodile.matplotlib_management import *
import crocodile.toolbox as tb
import numpy as np
import pandas as pd
import platform


console = Console()
pretty.install()

_ = f"Python {platform.python_version()} in VE `{os.getenv('VIRTUAL_ENV')}` On {platform.system()}."
_ = Text(_); _.stylize("bold blue")
console.rule(_, style="bold red", align="center")
# link to tutorial or github
_ = Text(f"Crocodile Shell")
_.stylize("#93e6c7 on #093006")
print(_, __import__('crocodile').__version__)
print("Made with 🐍 | Built with ❤️")

tb.D.set_numpy_display()
tb.D.set_pandas_display()
D = Display; L = List; E = Experimental; S = Struct


if platform.system() == "Windows":
    _1x = P.home().joinpath(r"AppData/Roaming/npm/figlet").exists()
    _2x = P.home().joinpath(r"AppData/Roaming/npm/lolcatjs").exists()
    _3x = P.home().joinpath(r"AppData/Local/Microsoft/WindowsApps/boxes.exe").exists()

    if _1x and _2x and _3x:
        if random.choice([True, True, False]):
            from crocodile.msc.ascii_art import FIGJS_FONTS  # , BoxStyles
            font = random.choice(FIGJS_FONTS)
            # print(f"{font}\n")
            box_style = random.choice(['whirly', 'xes', 'columns', 'parchment', 'scroll', 'scroll-akn', 'diamonds', 'headline', 'nuke', 'spring', 'stark1'])
            _cmd = f'figlet -f "{font}" "crocodile" | boxes -d "{box_style}" | lolcatjs'
            print(_cmd)
            os.system(_cmd)  # | lolcat
        else:
            from crocodile.msc.ascii_art import ArtLib
            # from rgbprint import gradient_scroll, Color
            # gradient_scroll(ArtLib.cowsay("crocodile"), start_color=0x4BBEE3, end_color=Color.medium_violet_red, times=3)
            _new_art = tb.P.temp().joinpath("tmp_arts").create().joinpath(f"{tb.randstr()}.txt")
            _new_art.write_text(ArtLib.cowsay("crocodile"))  # utf-8 encoding?
            os.system(f'type {_new_art} | lolcatjs')  # | lolcat
    else:
        print(f"Missing ascii art dependencies. Install with: iwr bit.ly/cfgasciiartwindows | iex")
        _default_art = P(__file__).parent.joinpath("art").search().sample(size=1)[0]
        print(_default_art.read_text())
else:
    # _x1 = P.home().joinpath(".nix-profile/bin/cowsay").exists()  # P(r"/usr/games/cowsay").exists()
    # _x2 = P.home().joinpath(".nix-profile/bin/lolcat").exists()  # P(r"/usr/games/lolcat").exists()
    # _x3 = P.home().joinpath(".nix-profile/bin/boxes").exists()  # P(r"/usr/bin/boxes").exists()
    # _x4 = P.home().joinpath(".nix-profile/bin/figlet").exists()  # P(r"/usr/bin/figlet").exists()
    def is_executable_in_path(executable_name):
        path_dirs = os.environ['PATH'].split(os.pathsep)
        for path_dir in path_dirs:
            path_to_executable = os.path.join(path_dir, executable_name)
            if os.path.isfile(path_to_executable) and os.access(path_to_executable, os.X_OK): return True
        return False
    _x1 = is_executable_in_path("cowsay")
    _x2 = is_executable_in_path("lolcat")
    _x3 = is_executable_in_path("boxes")
    _x4 = is_executable_in_path("figlet")

    if _x1 and _x2 and _x3 and _x4:
        _dynamic_art = random.choice([True, True, True, True, False])  # classic art (True) or boxes (False)
        if _dynamic_art:
            from crocodile.msc.ascii_art import get_art
            _new_art = tb.P.temp().joinpath("tmp_arts").create().joinpath(f"{tb.randstr()}.txt")
            get_art("crocodile", artlib=None, file=_new_art, verbose=False)
            os.system(f"cat {_new_art} | lolcat")  # full path since lolcat might not be in PATH.
        else:
            _default_art = P(__file__).parent.joinpath("art").search().sample(size=1)[0]
            print(_default_art.read_text())
    else:
        print(f"Missing ascii art dependencies. Install with: curl bit.ly/cfgasciiartlinux -L | sudo bash")
        _default_art = P(__file__).parent.joinpath("art").search().sample(size=1)[0]
        print(_default_art.read_text())
print("\n")


def build_parser():
    parser = argparse.ArgumentParser(description="Generic Parser to launch a script in a separate window.")
    parser.add_argument("--cmd", "-c", dest="cmd", help="Python command.", default="")
    parser.add_argument("--file", "-f", dest="file", help="Python command.", default="")

    args = parser.parse_args()
    # tb.Struct(args.__dict__).print(as_config=True)
    if args.file:
        code = P(args.file).read_text()
        print(code)
        exec(code, globals())
    elif args.cmd:
        print(args.cmd)
        exec(args.cmd, globals())


if __name__ == "__main__":
    build_parser()




import time
from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine, text, inspect
from sqlalchemy.sql.schema import MetaData
import crocodile.toolbox as tb
import pandas as pd


"""
https://docs.sqlalchemy.org/en/14/tutorial/index.html#a-note-on-the-future

"""

tb.Display.set_pandas_display()


class DBMS:
    """Implementation Philosophy:
    * Always use sqlalchemy API and avoid sql-dielect specific language.
    * Engine is provided externally. It is the end-user's business to make this engine.
    """
    def __init__(self, engine, db=None, sch=None, vws=False):
        self.eng = engine
        self.con = None
        self.ses = None
        self.insp = None
        self.meta = None
        self.path = tb.P(self.eng.url.database)

        self.db = db
        self.sch = sch
        self.vws = vws
        self.schema = None
        # self.tables = None
        # self.views = None
        self.sch_tab = None
        self.sch_vws = None
        self.refresh()

    def refresh(self, sch=None):  # fails if multiple schemas are there and None is specified
        self.con = self.eng.connect()
        self.ses = sessionmaker()(bind=self.eng)  # ORM style
        self.meta = MetaData()
        self.meta.reflect(bind=self.eng, schema=sch or self.sch)
        self.insp = inspect(subject=self.eng)
        self.schema = tb.L(self.insp.get_schema_names())
        self.sch_tab = tb.Struct.from_keys_values(self.schema, self.schema.apply(lambda x: self.insp.get_table_names(schema=x)))
        self.sch_vws = tb.Struct.from_keys_values(self.schema, self.schema.apply(lambda x: self.insp.get_view_names(schema=x)))
        return self

    def __getstate__(self): return tb.Struct(self.__dict__.copy()).delete(keys=["eng", "con", "ses", "insp", "meta"]).update(path=self.path.collapseuser(strict=False)).__dict__
    def __setstate__(self, state): self.__dict__.update(state); self.eng = self.make_sql_engine(self.path); self.refresh()

    @classmethod
    def from_local_db(cls, path=None, echo=False): return cls(engine=cls.make_sql_engine(path=path, echo=echo))
    def __repr__(self): return f"DataBase @ {self.eng}"
    def get_columns(self, table, sch=None): return self.meta.tables[self._get_table_identifier(table, sch)].exported_columns.keys()
    def close(self, sleep=2): print(f"Terminating database `{self.path.as_uri() if 'memory' not in self.path else self.path}`"); self.con.close(); self.ses.close(); self.eng.dispose(); time.sleep(sleep)
    def _get_table_identifier(self, table, sch):
        if sch is None: sch = self.sch
        if sch is not None: return sch + "." + table
        else: return table

    @staticmethod
    def make_sql_engine(path=None, echo=False, dialect="sqlite", driver=["pysqlite", "DBAPI"][0]):
        """Establish lazy initialization with database"""
        if path == "memory":
            print("Linking to in-memory database.")
            from sqlalchemy.pool import StaticPool  # see: https://docs.sqlalchemy.org/en/14/dialects/sqlite.html#using-a-memory-database-in-multiple-threads
            return create_engine(url=f"{dialect}+{driver}:///:memory:", echo=echo, future=True, poolclass=StaticPool, connect_args={"check_same_thread": False})
        path = tb.P.tmpfile(folder="tmp_dbs", suffix=".db") if path is None else tb.P(path).expanduser().absolute().create(parents_only=True)
        print(f"Linking to database at {path.as_uri()}")
        return create_engine(url=f"{dialect}+{driver}:///{path}", echo=echo, future=True)  # echo flag is just a short for the more formal way of logging sql commands.

    # ==================== QUERIES =====================================
    def execute_as_you_go(self, *commands, res_func=lambda x: x.all(), df=False):
        with self.eng.connect() as conn:
            for command in commands: result = conn.execute(text(command))
            conn.commit()  # if driver is sqlite3, the connection is autocommitting. # this commit is only needed in case of DBAPI driver.
        return res_func(result) if not df else pd.DataFrame(res_func(result))

    def execute_begin_once(self, command, res_func=lambda x: x.all(), df=False):
        with self.eng.begin() as conn:
            result = conn.execute(text(command))  # no need for commit regardless of driver
            result = res_func(result)
        return result if not df else pd.DataFrame(result)

    def execute(self, command, df=False):
        with self.eng.begin() as conn: result = conn.execute(text(command))
        return result if not df else pd.DataFrame(result)

    def execute_script(self, command, df=False):
        with self.eng.begin() as conn: result = conn.executescript(text(command))
        return result if not df else pd.DataFrame(result)

    # ========================== TABLES =====================================
    def read_table(self, table, sch=None, size=100):
        res = self.con.execute(text(f'''SELECT * FROM "{self._get_table_identifier(table, sch)}"'''))
        return pd.DataFrame(res.fetchmany(size))

    def insert_dicts(self, table, *mydicts):
        cmd = f"""INSERT INTO {table} VALUES """
        for mydict in mydicts: cmd += f"""({tuple(mydict)}), """
        self.execute_begin_once(cmd)

    def describe_table(self, table, sch=None, dtype=True):
        print(table.center(100, "="))
        self.refresh()
        tbl = self.meta.tables[table]
        count = self.ses.query(tbl).count()
        res = tb.Struct(name=table, count=count, size_mb=count * len(tbl.exported_columns) * 10 / 1e6)
        res.print(dtype=False, as_config=True)
        dat = self.read_table(table=table, sch=sch, size=2)
        cols = self.get_columns(table, sch=sch)
        df = pd.DataFrame.from_records(dat, columns=cols)
        print("SAMPLE:\n", df)
        if dtype: print("\nDETAILED COLUMNS:\n", pd.DataFrame(self.insp.get_columns(table)))
        print("\n" * 3)


if __name__ == '__main__':
    pass




# import os
import random
# import crocodile.toolbox as tb
import textwrap
import subprocess
from crocodile.core import install_n_import


# https://github.com/VaasuDevanS/cowsay-python
# https://github.com/pwaller/pyfiglet
# https://github.com/tehmaze/lolcat/
# https://github.com/wasi-master/gradient_figlet/

"""

"""


class ArtLib:
    @staticmethod
    def cowsay(text):
        cowsay = install_n_import("cowsay")
        char = random.choice(cowsay.char_names)
        return cowsay.get_output_string(char_name=char, text=text)

    @staticmethod
    def figlet(text):
        pyfiglet = install_n_import("pyfiglet")
        fs = pyfiglet.FigletFont.getFonts()
        f = pyfiglet.Figlet(font=random.choice(fs))
        return f.renderText(text)


class BoxStyles:
    language = ['ada-box', 'caml', 'boxquote', 'stone', 'tex-box', 'shell', 'simple' 'c', 'cc', 'html']
    scene = ['whirly', 'xes', 'columns', 'parchment', 'scroll', 'scroll-akn', 'diamonds', 'headline', 'nuke', 'spring', 'stark1']  # , 'important3'
    character = ['capgirl', 'cat', 'boy', 'girl', 'dog', 'mouse', 'santa', 'face', 'ian_jones', 'peek', 'unicornsay']


class CowStyles:
    eyes = ['-b', '-d', '-g', '-h', '-l', '-L', '-n', '-N', '-p', '-s', '-t', '-w', '-y']
    figures = ['apt', 'bunny', 'cheese', 'cock', 'cower', 'daemon', 'default', 'dragon',
               'dragon-and-cow', 'duck', 'elephant', 'elephant-in-snake', 'eyes', 'fox', 'ghostbusters',
               'gnu', 'kangaroo', 'kiss', 'milk',
               'moose', 'pony', 'pony-smaller', 'sheep', 'skeleton', 'snowman', 'stegosaurus',  # 'suse',
               'three-eyes', 'turkey', 'turtle', 'tux', 'unipony', 'unipony-smaller', 'vader', 'vader']  # 'hellokitty' 'mech-and-cow'  # 'moofasa', 'stimpy', 'calvin', , 'ren', 'koala', 'flaming-sheep' , 'bud-frogs' , 'kosh' , 'luke-koala'


FIGLET_FONTS = ['banner', 'big', 'standard']

FIGJS_FONTS = ['3D Diagonal', '3D-ASCII', '4Max', '5 Line Oblique', 'Acrobatic', 'ANSI Regular', 'ANSI Shadow', 'Avatar', 'Banner', 'Banner3-D', 'Banner4',
               'Basic', 'Big Money-ne', 'Big Money-nw', 'Big Money-se', 'Big Money-sw', 'Big', 'Bloody', 'Bolger', 'Braced', 'Bright',
               'DOS Rebel',
               'Elite', 'Epic', 'Flower Power',
               'Fraktur', # 'Isometric4'. 'AMC Tubes', 'Banner3', Alligator2
               'Star Wars',
               'Sub-Zero', 'The Edge', 'USA Flag', 'Varsity', "Doom"]  # too large  Crazy 'Sweet', 'Electronic', 'Swamp Land', Crawford, Alligator


def get_art(comment=None, artlib=None, style=None, super_style='scene', prefix='', file=None, verbose=True):
    """ takes in a comment and does the following wrangling:
    * text => figlet font => boxes => lolcat
    * text => cowsay => lolcat
    """
    if comment is None: comment = subprocess.run("fortune", shell=True, capture_output=True, text=True).stdout
    if artlib is None: artlib = random.choice(['boxes', 'cowsay'])
    to_file = '' if not file else f'> {file}'
    if artlib == 'boxes':
        if style is None: style = random.choice(BoxStyles.__dict__[super_style or random.choice(['language', 'scene', 'character'])])
        fonting = f'figlet -f {random.choice(FIGLET_FONTS)}'
        cmd = f"""echo "{comment}" | {fonting} | boxes -d {style} {to_file}"""
    else:
        if style is None: style = random.choice(CowStyles.figures)
        cmd = f"""echo "{comment}" | cowsay -f {style} {to_file}"""
    res = subprocess.run(cmd, text=True, capture_output=True, shell=True).stdout
    res = textwrap.indent(res, prefix=prefix)
    if verbose:
        print(f'Using style: {style} from {artlib}', '\n' * 3)
        print(cmd)
        print(res)
    return res


if __name__ == '__main__':
    pass




from crocodile.file_management import P, install_n_import, datetime
from crocodile.core import List
from crocodile.meta import Scheduler, Log
import time


def share(path):
    response = install_n_import("requests").post(url=r'https://file.io', data=dict(name=path.name, expires="2022-08-01", title="a descriptivefile title", maxDownloads=1, autoDelete=True, private=False, description="its a file, init?", ), files={'file': path.read_bytes()})
    return response.json()['link'] if ['link'] in response.json() else response.json()
def edit_video(path, t_start=0, t_end=None, speed=1, suffix=None, rotate=0, volume=1.0, fps=25):
    ed = install_n_import("moviepy").editor
    clip = ed.VideoFileClip(path); print(f"{clip.size=}, {clip.duration=}, {clip.fps=}")
    clip.subclip(t_start=t_start, t_end=t_end).rotate(rotate).volumex(volume).fx(ed.vfx.speedx, speed).write_videofile(path.append("_modified").with_suffix(path.suffix if suffix is None else suffix), fps=fps)


def capture_from_webcam(show=True, wait=True, save=False):
    cv2 = install_n_import("cv2", "opencv-python")
    cam = cv2.VideoCapture(0)
    cam.set(cv2.CAP_PROP_FRAME_WIDTH, 3000)
    cam.set(cv2.CAP_PROP_FRAME_HEIGHT, 2000)
    ret, frame = cam.read()
    while True and wait:
        ret, frame = cam.read()
        if show: cv2.imshow('frame', frame)
        if cv2.waitKey(1) & 0xFF == ord('q'): break
    cam.release()
    cv2.destroyAllWindows()
    if save:
        path = P.tmpfile(suffix=".jpg")
        cv2.imwrite(str(path), frame)
        return path
    return frame


def qr(txt): install_n_import("qrcode").make(txt).save((file := P.tmpfile(suffix=".png")).__str__()); return file()
def count_number_of_lines_of_code_in_repo(path=P.cwd(), extension=".py", r=True, **kwargs): return P(path).search(f"*{extension}", r=r, **kwargs).read_text(encoding="utf-8").splitlines().apply(len).np.sum()
def profile_memory(command): psutil = install_n_import("psutil"); before = psutil.virtual_memory(); exec(command); after = psutil.virtual_memory(); print(f"Memory used = {(after.used - before.used) / 1e6}")


class Cycle:
    def __init__(self, iterable=None, max_idx=None):
        if iterable is not None:
            self.list = iterable
            self.index, self.prev_index = 0, -1
        else:
            if max_idx is not None:
                self.list = list(range(max_idx))
                self.index, self.prev_index = 0, -1
            else:
                self.list = []
                self.index, self.prev_index = -1, -2
    def next(self): self.prev_index = self.index; self.index += 1; self.index = 0 if self.index >= len(self.list) else self.index; return self.list[self.index]
    def previous(self): self.prev_index = self.index; self.index -= 1; self.index = len(self.list) - 1 if self.index < 0 else self.index; return self.list[self.index]
    def set_value(self, value): self.prev_index = self.index; self.index = self.list.index(value)
    def get_value(self): return self.list[self.index]
    def get_index(self): return self.index
    def set_index(self, index): self.prev_index = self.index; self.index = index
    def expand(self, val=None): self.list += (val if type(val) is list else ([self.list[-1] + 1]) if len(self.list) else [0])
    def __add__(self, other): pass  # see behviour of matplotlib cyclers.
    def __repr__(self): return f"Cycler @ {self.index}: {self.list[self.index]}"
class DictCycle(Cycle):
    def __init__(self, strct): super(DictCycle, self).__init__(iterable=strct.items()); self.keys = strct.keys()
    def set_key(self, key): self.index = list(self.keys).index(key)


def tree(self, level: int = -1, limit_to_directories: bool = False, length_limit: int = 1000, stats=False, desc=None):
    # Based on: https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python"""
    space, branch, tee, last, dir_path, files, directories = '    ', '│   ', '├── ', '└── ', self, 0, 0
    def get_stats(apath): return (f" {(sts := apath.stats(printit=False)).size} MB. {sts.content_mod_time}. " + desc(apath) if desc is not None else "") if stats or desc else ""
    def inner(apath: P, prefix: str = '', level_=-1):
        nonlocal files, directories
        if not level_: return  # 0, stop iterating
        pointers = [tee] * (len(content := apath.search("*", files=not limit_to_directories)) - 1) + [last]
        for pointer, path in zip(pointers, content):
            if path.is_dir():
                yield prefix + pointer + path.name + get_stats(path)
                directories, extension = directories + 1, branch if pointer == tee else space
                yield from inner(path, prefix=prefix + extension, level_=level_ - 1)
            elif not limit_to_directories: yield prefix + pointer + path.name + get_stats(path); files += 1
    print(dir_path.name)
    iterator = inner(dir_path, level_=level)
    [print(line) for line in __import__("itertools").islice(iterator, length_limit)]
    print(f'... length_limit, {length_limit}, reached, counted:') if next(iterator, None) else None
    print(f'\n{directories} directories' + (f', {files} files' if files else ''))


def get_compressable_directories(path, max_size_mb=15_000):
    tmp_results = path.search("*", r=False)
    dirs2compress = List()
    dirs_violating = List()
    files_violating = List()
    for item in tmp_results:
        if item.size() > max_size_mb:  # should be parsed
            if item.is_file():
                print(f"File `{item}` has size larger than maximum allowed, consider manual handling")
                files_violating.append(item)
            else:
                tmp_dirs2compress, tmp_dirs_violating, tmp_files_violating = get_compressable_directories(item, max_size_mb=max_size_mb)
                if len(tmp_dirs2compress) == 0:
                    print(f"Directory `{item}` has size larger than maximum allowed, but when parsed for subdirectories to be compressed, nothing could be found. Handle manually")
                    dirs_violating.append(item)
                dirs2compress += tmp_dirs2compress
                dirs_violating += tmp_dirs_violating
                files_violating += tmp_files_violating
        else:
            if item.is_dir() and len(item.search("*", folders=False, r=True)) > 10:  # too many files ==> compress
                dirs2compress.append(item)
    return dirs2compress, dirs_violating, files_violating


if __name__ == '__main__':
    pass



from setuptools import setup
from myresources.crocodile import __version__
import setuptools

with open("README.md", "r", encoding="utf-8") as file:
    long_desc = file.read()  # to help adding README to PyPi website not only Github

# get python modules:
# tb.P("./myresources").search("*.py", r=True).apply(lambda x: x.split(at="myresources", sep=-1)[1].as_posix().replace("/", ".")[:-3]).filter(lambda x: "__init__" not in x).list

setup(
    name='crocodile',
    version=__version__,
    packages=['crocodile'],
    # packages=setuptools.find_packages(where="myresources"),
    package_dir={'': 'myresources'},
    # py_modules=setuptools.find_packages(where="myresources") + ['crocodile.msc.ascii_art'],
    py_modules=['crocodile.core',
                'crocodile.croshell',
                'crocodile.database',
                'crocodile.deeplearning',
                'crocodile.deeplearning_torch',
                'crocodile.environment',
                'crocodile.file_management',
                'crocodile.matplotlib_management',
                'crocodile.meta',
                'crocodile.plotly_management',
                'crocodile.run',
                'crocodile.toolbox',
                'crocodile.cluster.data_transfer',
                'crocodile.cluster.distribute',
                'crocodile.cluster.meta_handling',
                'crocodile.cluster.remote_machine',
                'crocodile.cluster.script_execution',
                'crocodile.cluster.script_notify_upon_completion',
                'crocodile.cluster.trial_file',
                'crocodile.comms.gdrive',
                'crocodile.comms.helper_funcs',
                'crocodile.comms.notification',
                'crocodile.comms.onedrive',
                'crocodile.msc.ascii_art',
                'crocodile.msc.dl_template',
                'crocodile.msc.numerical',
                'crocodile.msc.odds']
    ,
    url='https://github.com/thisismygitrepo/crocodile',
    project_urls={
        "Bug Tracker": "https://github.com/thisismygitrepo/crocodile/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",  # 3.9
    license='Apache 2.0',
    author='Alex Al-Saffar',
    author_email='programmer@usa.com',
    description='Making Python even more convenient by extending list and dict and pathlib and more.',
    long_description=long_desc,
    long_description_content_type="text/markdown",
    package_data={'crocodile': ['art/*']},
    # include_package_data=True,
    install_requires=[
        # CORE:
        "numpy",
        # "scipy",  # heavy-weight.
        "pandas",
        "matplotlib",
        "tabulate",  # pretty printing of tables

        # Accessories
        "joblib",  # multitasking
        "ipython",  # interactive python
        "fire",  # for automatic CLI interface
        "tqdm",  # for progress bar
        "dill",  # extends pickle
        "cryptography",  # for encoding
        "paramiko",  # for SSH
        "requests",  # interacting with web
        "cycler",

        # Developer Tools
        "setuptools",  # for packaging  # correct place is in .toml
        "wheel",
        "twine",  # for pushing package to pypi.org
        "pytest",
        # popular alternative to builtint unittest  # consider splitting requirements to development and production versions.
        # "scikit-image",  # image processing. Heavy-weight.
        # torch
        # tensorflow
    ]

)

with open("./myresources/crocodile/art/happy_croco", "r") as file:
    croco = file.read()  # search ascii art or characters art.
print(croco)




# import numpy as np
# import matplotlib.pyplot as plt
import crocodile.toolbox as tb
import smtplib
import imaplib
# from email import message
# from email import encoders
# from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart


"""

"""


def get_gtihub_markdown_css(): return tb.P(r'https://raw.githubusercontent.com/sindresorhus/github-markdown-css/main/github-markdown-dark.css').download(memory=True).text


class Email:
    @staticmethod
    def get_source_of_truth(): return tb.P.home().joinpath("dotfiles/creds/msc/source_of_truth.py").readit(strict=False)

    def __init__(self, config):
        self.config = config
        self.server = smtplib.SMTP(host=self.config["smpt_host"], port=self.config["smpt_port"])
        self.server.login(self.config['email_add'], password=self.config["get_password"]())

    def send_message(self, to, subject, body, txt_to_html=True, attachments=None):
        body += "\n\nThis is an automated email sent via crocodile.comms script."
        # msg = message.EmailMessage()
        msg = MIMEMultipart("alternative")
        msg["subject"] = subject
        msg["From"] = self.config['email_add']
        msg["To"] = to
        # msg['Content-Type'] = "text/html"
        # msg.set_content(body)

        # <link rel="stylesheet" href="github-markdown.css">
        # <link type="text/css" rel="stylesheet" href="https://raw.githubusercontent.com/sindresorhus/github-markdown-css/main/github-markdown-dark.css" />

        if txt_to_html:  # add gtihub markdown stylesheet
            body = f"""
<!DOCTYPE html>
<html>
<meta name="viewport" content="width=device-width, initial-scale=1">
<style>
{get_gtihub_markdown_css()}
    .markdown-body {{
        box-sizing: border-box;
        min-width: 200px;
        max-width: 980px;
        margin: 0 auto;
        padding: 45px;
        line-height: 1.8;
    }}
    @media (max-width: 767px) {{.markdown-body {{padding: 15px;}}
    }}
</style>
<body>
<div class="markdown-body">
{tb.install_n_import("markdown").markdown(body)}
</div>
</body>
</html>"""
        msg.attach(MIMEText(body, "html"))
        # if attachments is None: attachments = []  # see: https://fedingo.com/how-to-send-html-mail-with-attachment-using-python/
        # for attachment in attachmenthrs: msg.attach(attachment.read_bytes(), filename=attachment.stem, maintype="image", subtype=attachment.suffix)
        # for attachment in attachments: msg.attach(attachment.read_bytes(), filename=attachment.stem, maintype="application", subtype="octet-stream")

        self.server.send_message(msg)

    @staticmethod
    def manage_folders(email_add, pwd):
        server = imaplib.IMAP4()
        server.starttls()
        server.login(email_add, password=pwd)

    def send_email(self, to_addrs, msg): return self.server.sendmail(from_addr=self.config['email_add'], to_addrs=to_addrs, msg=msg)
    def close(self): self.server.quit()    # Closing is vital as many servers do not allow mutiple connections.

    @staticmethod
    def send_and_close(config_name, to, subject, msg): tmp = Email(config=Email.get_source_of_truth().EMAIL[config_name]); tmp.send_message(to, subject, msg); tmp.close()


class PhoneNotification:  # security concerns: avoid using this.
    def __init__(self, bulletpoint_token):
        pushbullet = tb.install_n_import("pushbullet")
        self.api = pushbullet.Pushbullet(bulletpoint_token)
    def send_notification(self, title="Note From Python", body="A notfication"): self.api.push_note(title=title, body=body)
    @staticmethod
    def open_website(): tb.P(r"https://www.pushbullet.com/").readit()
    @staticmethod  # https://www.youtube.com/watch?v=tbzPcKRZlHg
    def try_me(bulletpoint_token): n = PhoneNotification(bulletpoint_token); n.send_notification()


if __name__ == '__main__':
    pass




__version__ = "10.4.0"
__author__ = "Alex Al-Saffar"
__date__ = "2023-02-24"
docs = r"https://crocodile.readthedocs.io/en/latest/"
git = r"https://github.com/thisismygitrepo/crocodile"
pypi = r"https://pypi.org/project/crocodile"
release_notes = """
added bzip2 compression
"""

